#!/usr/bin/perl -w
#
#
#	"Deus Ex Machina" 
#	
#	ICFP 2002 Competition entry by team Superficial Intelligence Labs (Lightning Yearners)  [SILLY]
#
#	Contact email: frode@freenix.no
#
#	To run, do './dem.pl host port'
#	To run faster, do './dem.pl host port > /dev/null'
#
#	GOTOs have been used liberally. :)
#

use IO::Socket;

my ($hostname, $portnum, $sock);

my ($boardx, $boardy, @board);

my (@boardsplength, @boardspdir);

my ($botid, $botcapacity, $botcashmax);	# initial state
my ($botcurrcap, $botcurrcash);			# running state

my ($roundnum);

# other bots
my (%bots);		# key: bot ID. value: (reference) hash:
					#  'posx' => last known position
					#  'posy'
					#  'pkgs' => (reference) hash of (references) pkg hashes carried.

my (%pkg_onsite);
my (%pkg_known);	# key: package id. value (reference) hash	

					#	'posx' => last known (floor) position
					#   'posy'
					#	'targetx' => known delivery location
					#	'targety'
					#   'weight' => weight
					#   'owner' = botid or undef.

my ($mission);
					# 0 - roaming mindlessly
					# 1 - get package
					# 2 - deliver package
					# 3 - kill
					#		botid - bot to nuke

my ($msntargetx, $msntargety, $msntargetbot);
my ($prevtargetx, $prevtargety);		

my ($str, $i, $j);

$| = 0;

$hostname = shift;
$portnum = shift;

if (!$hostname || !$portnum) {
	die("Specify hostname and portnum");
}

$sock = IO::Socket::INET->new(PeerAddr => $hostname,
							 PeerPort => $portnum,
							 Proto => 'tcp');

die "couldn't connect $!" unless $sock;

$sock->autoflush();


# do some init stuff
$roundnum = 0;
$mission = 0;
$prevtargetx = $prevtargety = 0;
$msntargetx = $msntargety = 0;

#    1.  Client: send the string Player  
print $sock "Player\n";



#    2. Server: send board
#		1. Server: board dimensions
$str = <$sock>;
if (!($str =~ /^(\d+)\s(\d+)/)) {
	die ("Couldn't parse board size [$str]");
}

($boardx, $boardy) = ($1, $2);

print  "board x: $boardx  board y: $boardy\n";


for ($i=0; $i <= $boardy+1 ; $i++) {
	for ($j=0; $j <= $boardx+1 ; $j++) {
		$board[$i][$j] = '#';
	}
}


#		2. Server: board row 1 ...
#		3. Server: ... other rows ...
#		4. Server: board row n 

for ($i=1; $i <= $boardy; $i++) {
	$str = <$sock>;
	for ($j=1; $j <= $boardx; $j++) {
		$board[$i][$j] = substr($str, ($j-1), 1);
	}
}


#    3. Server: send player's configuration
$str = <$sock>;
#			my ($botid, $botcapacity, $botcashmax);	# initial state
if (!($str =~ /^(\d+)\s(\d+)\s(\d+)/)) {
	die ("Couldn't parse initial player config [$str]");
}

($botid, $botcapacity, $botcashmax) = ($1, $2, $3);
($botcurrcap, $botcurrcash) = ($botcapacity, $botcashmax);

print  "botid: $botid\tbotcap: $botcurrcap\tbotcash: $botcurrcash\n";

#    4. Server: send initial server response (with the initial location of all players) 

readserverstate();
printboardstateannotated();











while ($botcurrcash > 0) {
	my ($roundbid, $roundaction, @picklist);
	my ($bx, $by);
	my ($pkgid);

	# keep going while we still have cash
	# (what if we're dead?)

	$roundbid = 1;
	$roundaction = "Move N";
	@picklist = ();
	@droplist = ();
	
	readpackagelist();

	($bx, $by) = ($bots{$botid}->{'posx'}, $bots{$botid}->{'posy'});

	# do something...

	# see if this is the delivery spot of something we're carrying

	foreach $pkgid (keys %{($bots{$botid}->{'pkgs'})}) {
		if ( ($pkg_known{$pkgid}->{'targetx'} == $bx) && ($pkg_known{$pkgid}->{'targety'} == $by)) {
			print  "Arrived at destination for package $pkgid!\n";
			push @droplist, $pkgid;
		}
	}

	if (scalar(@droplist) > 0) {
		$roundaction = "Drop " . join(" ", @droplist);
		# XXX: increase bid if robots are near.
		goto JUSTDOIT
	}

	# look for package onsite
	my $tempbotcurrcap = $botcurrcap;

	foreach $pkgid (keys %pkg_onsite) {
		if ($pkg_onsite{$pkgid}->{'weight'} < $tempbotcurrcap) {
			# pick up this one.
			push @picklist, $pkgid;
			$tempbotcurrcap -= $pkg_onsite{$pkgid}->{'weight'};
			print  "Adding pkgid $pkgid to picklist, temp currcap is now [$tempbotcurrcap]\n";
		}
	}

	if (scalar(@picklist) > 0) {
		$roundaction = "Pick " . join(" ", @picklist);
		# XXX: increase bid if robots are near.
		goto JUSTDOIT
	}


RESTART_LOOP:

	# maybe pick up packages if there are any.
	if (($mission == 0) || ($mission == 2 && (scalar(keys %{($bots{$botid}->{'pkgs'})})) == 0)) {
		# nothing to do
		# look for a known package..

		$mission = 1;

#		if (scalar(keys %pkg_known) > 0) {
#			# go fetch a known package.
#			print  "Going for known package\n";
#
#		} else {

#	XXX: should remove dropped (delivered) packages from pkg_known
#		 how to know that for 3rd party bots dropping?
		my ($homex, $homey);

			{
			$i = 1; $j = 1;
			FINDANOTHERHOMEBASE:

			# go to homebase
			print  "Looking for homebase\n";
			for (;$i<=$boardy;$i++) {
				for (; $j<=$boardx;$j++) {
					if ($board[$i][$j] eq '@') {
						($homey, $homex) = ($i, $j);
						goto FOUNDHOMEBASE;
					}
				}
				$j = 1;
			}

			# no home base..
			# go around kill people!
			print  "No more homebases to visit...\n";
			$mission = 3;
			$msntargetbot = 0;
			goto KILLABOT;

			FOUNDHOMEBASE:
			print  "New target: homebase at [$homey $homex]\n";
			($msntargetx, $msntargety) = ($homex, $homey);
		}


	}

	if ($mission == 2) {
		# deliver package
		# find a target...
			$pkg = (keys %{$bots{$botid}->{'pkgs'}})[0];
			print  "NOW MUST DELIVER PKG $pkg\n";
			($msntargetx, $msntargety) = ($bots{$botid}->{'pkgs'}->{$pkg}->{'targetx'}, $bots{$botid}->{'pkgs'}->{$pkg}->{'targety'} );

			print  "Target: " . $msntargety . ",". $msntargetx ."\n";

	}

KILLABOT:
	if ($mission == 3) {
		# kill. kill. kill.
		print  "    **** GOING POSTAL! **** \n";

		my ($cbotid, $seen);
		foreach $cbotid (keys %bots) {
			my $seen = 0;
			if ($cbotid == $msntargetbot) {
					$seen = 1;
			}
		}
		
		if ((!$seen) || ($msntargetbot == 0)) {
			# find another bot.
			# XXX: Don't kill ourselves ;)
			my $t = 0;
			while (($msntargetbot == $botid) || ($msntargetbot == 0)) {
				$msntargetbot = (keys %bots)[$t++];
				if ($t > scalar(keys %bots)) {
					# no bots either.
					# well then. 
					# spend all our cash and leave!
					
					#print  "  **** No more fun to do, spending all the cash!\n";
					#$roundbid = $botcurrcash;
					#$roundaction = "Drop 0";
					#goto JUSTDOIT;

					# on second thought, just cycle.
					$mission = 0;
					goto RESTART_LOOP;


				}
			}
			print  " Retargeting kill target to bot [$msntargetbot]\n";
		}


		# go after the bot..
		($msntargetx, $msntargety) = ($bots{$msntargetbot}->{'posx'}, $bots{$msntargetbot}->{'posy'});
		# XXX: This is slow and will regenerate shortest path every time.
		#	Optimisation: Bot moves at most 1, something clever should be done here..

	}


	if ($msntargetx != $prevtargetx || $msntargety != $prevtargety) {
		my @scannext;
		my $c;
		# recalc shortest path.
		# init values
		for ($i=0; $i<$boardy+1; $i++) {
			for ($j=0; $j<$boardx+1; $j++) {
				$boardsplength[$i][$j] = (2**31);	# "infinite"
				$boardspdir[$i][$j] = ' ';	# "infinite"
			}
		}

		($prevtargetx, $prevtargety) = ($msntargetx, $msntargety);
		
		$boardsplength[$msntargety][$msntargetx] = 0;
		$boardspdir[$msntargety][$msntargetx] = 0;

		push (@scannext, [($msntargety, $msntargetx)]);
		
		print  "Regenerating shortest path...\n";



		my $scaniter = 0;


		while (scalar(@scannext)) {
			$scaniter++;
			$nextpos = pop (@scannext);
			my ($y, $x) = ($nextpos->[0], $nextpos->[1]);

#	This doesn't work, of course, it finds a (very) sub optimal path...
#			if (($x == $bx) && ($y == $by)) {
#				# stop now, we've found full path.
#				print  "Found path after $scaniter iterations\n";
#				goto PATHFOUND
#			}

#			print  "Scanning neighbours of [$y,$x] distance [" . $boardsplength[$y][$x]  . "]  dir [". $boardspdir[$y][$x] ."]\n";
			my ($curcost) = $boardsplength[$y][$x]  + 1;

			$c = $board[$y-1][$x]; 
			if ($c eq '.' || $c eq '@') {
				if ($boardsplength[$y-1][$x] > $curcost) {
					$boardsplength[$y-1][$x] = $curcost;
					$boardspdir[$y-1][$x] = 'N';
					push (@scannext, [($y-1, $x)]);
				}
			}

			$c = $board[$y+1][$x]; 
			if ($c eq '.' || $c eq '@') {
				if ($boardsplength[$y+1][$x] > $curcost) {
					$boardsplength[$y+1][$x] = $curcost;
					$boardspdir[$y+1][$x] = 'S';
					push (@scannext, [($y+1, $x)]);
				}
			}

			$c = $board[$y][$x-1]; 
			if ($c eq '.' || $c eq '@') {
				if ($boardsplength[$y][$x-1] > $curcost) {
					$boardsplength[$y][$x-1] = $curcost;
					$boardspdir[$y][$x-1] = 'E';
					push (@scannext, [($y, $x-1)]);
				}
			}

			$c = $board[$y][$x+1]; 
			if ($c eq '.' || $c eq '@') {
				if ($boardsplength[$y][$x+1] > $curcost) {
					$boardsplength[$y][$x+1] = $curcost;
					$boardspdir[$y][$x+1] = 'W';
					push (@scannext, [($y, $x+1)]);
				}
			}
		}

	}
	PATHFOUND:

	if (($mission == 1) || ($mission == 2)) {
		# go to target

		if ($boardspdir[$by][$bx] eq '0') {
			if ($mission == 1) {
				# ok, we've arrived at the target but
				# obviously there's no package here anymore (or we'd picked it up and GOTO JUSTDOIT)

				# OR all the packages are too big.
				# in any case, we need to do something else.

				# retarget a different home base.
				
				# XXX add code
				# $i, $j are $y, $x coords to start searching.
				print  " ** nothing to do at target base, retargeting\n";
				($i, $j) = ($msntargety, $msntargetx+1);
				goto FINDANOTHERHOMEBASE;

			}
			else {
				# we're at target and for some reason (bug) we haven't done what we're supposed to do.
				# never issues a Move 0, that kills us.
				# do something else.
				print  " ** nothing to do at target base (bug), going postal\n";
				$mission = 3;
				$msntargetbot = 0;
				goto KILLABOT;

			}
		}


		$roundaction = "Move " . $boardspdir[$by][$bx];


#		if ($msntargetx > $bx) {
#			# go east
#			$roundaction = "Move E";
#		} elsif ($msntargetx < $bx) {
#			# go west
#			$roundaction = "Move W";
#		} elsif ($msntargety > $by) {
#			# go north (increase Y)
#			$roundaction = "Move N";
#		} elsif ($msntargety < $by) {
#			# go south
#			$roundaction = "Move S";
#		} 


	}

	

	JUSTDOIT:
	
	$botcurrcash -= $roundbid;

	# tell the server
	print  "******ACTION : " . $roundbid . " " . $roundaction . "\n";
	print $sock   $roundbid . " " . $roundaction . "\n";


	$roundnum++;
	readserverstate();
	printboardstateannotated();
}









sub readpackagelist {
	my ($str, @info);
	$str = <$sock>;
	die ("Couldn't read any more from socket $!") unless ($str) ;
	chomp ($str);

	%pkg_onsite = ();

	@info = split (/ /, $str);

	while (scalar(@info)>3) {
		my $pkgid = shift @info;
		my $targetx = shift @info;
		my $targety = shift @info;
		my $weight = shift @info;

		my $pkgdata;

		$pkgdata = $pkg_known{$pkgid};
		$pkgdata = {} unless defined($pkgdata);
		$pkgdata->{'posx'} = $bots{$botid}->{'posx'};	# our location
		$pkgdata->{'posy'} = $bots{$botid}->{'posy'};	# our location
		$pkgdata->{'targetx'} = $targetx;	
		$pkgdata->{'targety'} = $targety;
		$pkgdata->{'weight'} = $weight;
		$pkgdata->{'owner'} = undef;
		$pkg_known{$pkgid} = $pkgdata;

#		print  "Found package! id [$pkgid]  target [$targetx $targety] weight [$weight]";
		print  "pkg[$pkgid]";
		$pkg_onsite{$pkgid} = $pkgdata;
	}
	print  "\n";
}

sub readserverstate {
	my ($str, $curinf);
	my (@info, @infx);
	my (@botsseenthisround);

	$str = <$sock>;

	die ("Couldn't read any more from socket $!") unless ($str) ;

	chomp $str;

#	print  "Info for round $roundnum\t[$str]\n";

	@info = split(/\#/, $str);
	
	shift @info;	# throw away first item in @info as it is empty (string starts with #)

	@botsseenthisround = ();

	foreach $curinf (@info) {

		my $cbotid;
		my %cbotdata;


#		print  "Looking at bot data\t[$curinf]\n";
		@infx = split(/ /, $curinf);
		# @infx holds all operations on this bot.

		# get bot id
		$cbotid = $infx[0];
#		print  " Bot id $cbotid\n";

		push @botsseenthisround , $cbotid;

		# update bot data
		if (!defined($bots{$cbotid})) {
			# new bot...
			print  "    ** NEW BOT [$cbotid]\n";
			$cbotdata = {};
			$cbotdata->{'posx'} = 0;
			$cbotdata->{'posy'} = 0;
			$cbotdata->{'pkgs'} = {};
		} else {
			$cbotdata = $bots{$cbotid};
		}
		
		#	what happens to bots that don't do anything (are rebooted/pushed and stuck?)
		while (scalar(@infx) > 0) {
			# more info available on bot.
			my $opcode = shift @infx;

			if ($opcode eq 'N') {
				$cbotdata->{'posy'} += 1;		# north is increasing y position
				print  " Bot [$cbotid] walked north\n";
			} 
			elsif ($opcode eq 'S') {
				$cbotdata->{'posy'} -= 1;		
				print  " Bot [$cbotid] walked south\n";
			}
			elsif ($opcode eq 'E') {
				$cbotdata->{'posx'} += 1;		
				print  " Bot [$cbotid] walked east\n";
			}
			elsif ($opcode eq 'W') {
				$cbotdata->{'posx'} -= 1;		
				print  " Bot [$cbotid] walked west\n";
			}
			elsif ($opcode eq 'P') {
				# pick up package with ID...
				my $pkgid = shift @infx;
				
				# add to known package list..
				my $pkgdata;
				$pkgdata = $pkg_known{$pkgid};
				$pkgdata = {} unless defined($pkgdata);
				$pkgdata->{'posx'} = $cbotdata->{'posx'};
				$pkgdata->{'posy'} = $cbotdata->{'posy'};
				$pkgdata->{'owner'} = $cbotid;
				$pkg_known{$pkgid} = $pkgdata;

				$cbotdata->{'pkgs'}->{$pkgid} = $pkgdata;

				my $cbotpkgs = $cbotdata->{'pkgs'};
				$cbotpkgs->{$pkgid} = $pkgdata;

				# if it is us that picked up a package, it's time to change mode.
				if ($cbotid == $botid) {
					$mission = 2;
					($msntargetx, $msntargety) = ( $pkg_known{$pkgid}->{'targetx'},  $pkg_known{$pkgid}->{'targety'});

					$botcurrcap -= $pkg_known{$pkgid}->{'weight'};

					print  "We picked up package! Changing mission to deliver. Target ["
						.$pkg_known{$pkgid}->{'targetx'} . "," 
						.$pkg_known{$pkgid}->{'targety'} . "]\n";
				}

				print  " Bot [$cbotid] picked up package with id [$pkgid]\n";

			}
			elsif ($opcode eq 'D') {
				# drop package with ID...
				my $pkgid = shift @infx;

				# add to known package list..
				$pkgdata = $pkg_known{$pkgid};
				$pkgdata = {} unless defined($pkgdata);
				$pkgdata->{'posx'} = $cbotdata->{'posx'};
				$pkgdata->{'posy'} = $cbotdata->{'posy'};
				$pkgdata->{'owner'} = undef;
				$pkg_known{$pkgid} = $pkgdata;

				if ($cbotid == $botid) {
					$botcurrcap += $pkg_known{$pkgid}->{'weight'};
				}

				my $cbotpkgs = $cbotdata->{'pkgs'};
				delete ($cbotpkgs->{$pkgid});



				print  " Bot [$cbotid] dropped package with id [$pkgid]\n";
			}
			elsif ($opcode eq 'X') {
				my $pos = shift @infx;
				print  " Bot [$cbotid] set X position to [$pos]\n";
				$cbotdata->{'posx'} = $pos;
			}
			elsif ($opcode eq 'Y') {
				my $pos = shift @infx;
				print  " Bot [$cbotid] set Y position to [$pos]\n";
				$cbotdata->{'posy'} = $pos;
			}			
		}
	
		$bots{$cbotid} = $cbotdata;

	}

	# remove bots not seen
	foreach $cbotid (keys %bots) {
		my $seen = 0;
		my $cbotid2;
		foreach $cbotid2 (@botsseenthisround) {
			if ($cbotid2 == $cbotid) {
				$seen = 1;
			}
		}
		if (!$seen) {
			print  " Removing unseen bot $cbotid\n";
			delete $bots{$cbotid};
		}
	}


}



sub printboardstateannotated {
	my ($i, $j);
	my (@boardann);

	for ($i=0; $i <= $boardy+1; $i++) {
		for ($j=0; $j <= $boardx+1; $j++) {
			$boardann[$i][$j] = $board[$i][$j];
		}
	}

	
	print  "\n\n --- round [$roundnum] ---\n";
	print  "CASH [$botcurrcash]  FREE CAP [$botcurrcap] #PKGS [" . (keys %{($bots{$botid}->{'pkgs'})}) . "] MISSION [$mission] TARGET [" . $msntargety .",". $msntargetx ."]\n";

	foreach $_botid (keys %bots) {
		my $cbot = $bots{$_botid} ;
		if ($_botid == $botid) {
			# ourselves!
			$boardann [ $cbot->{'posy'} ] [ $cbot->{'posx'} ] = 'B';
		}
		else {
			$boardann [ $cbot->{'posy'} ] [ $cbot->{'posx'} ] = 'b';
		}
	}

	# print packages as well..
	# XXX: add this

	for ($i=0; $i <= $boardy+1; $i++) {
		for ($j=0; $j <= $boardx+1; $j++) {
			print  $boardann [$i][$j];
		}
		print  "\n";
	}

}
